/*******************************************************************************
 mod_smooth_streaming.c - An Apache module for streaming Quicktime/MPEG4 files.

 Copyright (C) 2007-2023 CodeShop B.V.

 For licensing see the LICENSE file
******************************************************************************/

#include <httpd.h>
#include <http_core.h>
#include <http_config.h>
#include <http_protocol.h>
#include <http_log.h>
#include <http_request.h> // for ap_update_mtime
#include <apr_version.h>
#include <apr_strings.h>  // for apr_itoa, apr_pstrcat
#include <apr_buckets.h>
#include <apr_date.h>
#include <ap_regex.h>

#ifdef _WIN32
#define snprintf _snprintf
#endif

#include <mod_streaming_export.h>
#include <mp4_process.h>
#include <mp4_rewrite.h>
#include <mp4_error.h>
#include <moov.h>
#include <mp4_options.h>
#include <output_bucket.h>
#include <mp4_handler_io.h>
#include <post_handler.h>
#include <s3_util.h>

#include <ap_log.h>
#include <apr_buckets_usp.h>
#include <subrequest.h>
#define SMOOTH_STREAMING_HANDLER "smooth-streaming.extensions"

#define STRINGIFY(a) STRINGIFY2(a)
#define STRINGIFY2(a) #a

#define MODULE_VERSION STRINGIFY(MOD_SMOOTH_STREAMING_VERSION_MAJOR) "." \
                       STRINGIFY(MOD_SMOOTH_STREAMING_VERSION_MINOR) "." \
                       STRINGIFY(MOD_SMOOTH_STREAMING_VERSION_REVISION)

#if 0

/* Mod-Smooth-Streaming configuration

[httpd.conf]
LoadModule smooth_streaming_module /usr/lib/apache2/modules/mod_smooth_streaming.so
AddHandler smooth-streaming.extensions .ism .isml

*/
#endif

typedef struct
{
  char const* fake;
  char const* real;
  int in_location;
} proxy_alias_t;

typedef struct
{
  apr_array_header_t* aliases;
  mp4_global_context_t* gctx;
  char const* license;
} server_config_t;

typedef struct
{
  int usp_enable_mmap;
  int usp_enable_subreq;
  int usp_prefer_static;
  int usp_iss_pass_through;
  int usp_dynamic_time_shift_buffer_depth;
  int usp_skip_rewrite;
  int usp_handle_ism;
  int usp_handle_f4f;
  int usp_handle_api;

  int fmp4_iss_fragment_not_found;
  int fmp4_iss_fragment_not_available;

  int fmp4_hds_fragment_not_found;
  int fmp4_hds_fragment_not_available;

  int fmp4_hls_fragment_not_found;
  int fmp4_hls_fragment_not_available;

  int fmp4_mpd_fragment_not_found;
  int fmp4_mpd_fragment_not_available;

  char const* s3_secret_key;
  char const* s3_access_key;
  char const* s3_region;
  char const* s3_security_token;
  int s3_use_headers;

  char const* transcode_proxy_pass;

  ap_expr_info_t const* usp_output_filter;

  int usp_force_content_length;
} dir_config_t;

static void* create_server_config(apr_pool_t* pool, server_rec* svr);
static void* merge_server_config(apr_pool_t* pool, void* basev, void* addv);
static void* create_dir_config(apr_pool_t* pool, char* dummy);
static void* merge_dir_config(apr_pool_t* pool, void* basev, void* addv);

static char const* set_usp_license_key(cmd_parms* cmd, void* cfg, char const* arg);
static char const* set_ism_proxy_pass(cmd_parms* cmd, void* cfg, char const* arg1, char const* arg2);
static char const* set_output_filter(cmd_parms* cmd, void* cfg, char const* arg);

static void register_hooks(apr_pool_t* p);

typedef struct
{
  char const* name;
  int count;
} deprecated_option_t;

static deprecated_option_t deprecated_options[] =
{
  { "Fmp4IssFragmentMissing", 0 },
  { "Fmp4HdsFragmentMissing", 0 },
  { "Fmp4HlsFragmentMissing", 0 },
  { "Fmp4MpdFragmentMissing", 0 }
};

static char const* deprecated_option(cmd_parms* cmd, void* cfg, char const* arg)
{
  int i;
  for(i = 0; i < sizeof(deprecated_options)/sizeof(deprecated_options[0]); ++i)
  {
    if(strcmp(cmd->cmd->name, deprecated_options[i].name) == 0)
    {
      ++deprecated_options[i].count;
    }
  }
  return NULL;
}

static apr_status_t cleanup_libfmp4(void* context);

static const command_rec smooth_streaming_cmds[] =
{
  AP_INIT_TAKE1("UspLicenseKey", set_usp_license_key,
    NULL,
    RSRC_CONF|ACCESS_CONF,
    "Set to USP license key"),

  AP_INIT_TAKE12("IsmProxyPass",
    set_ism_proxy_pass,
    NULL,
    RSRC_CONF|ACCESS_CONF,
    "Pass .ism to proxy (a virtual path and url)"),

  AP_INIT_TAKE1("S3SecretKey",
    ap_set_string_slot,
    (void*)APR_OFFSETOF(dir_config_t, s3_secret_key),
    ACCESS_CONF,
    "Pass the Amazon S3 Secret Key"),

  AP_INIT_TAKE1("S3AccessKey",
    ap_set_string_slot,
    (void*)APR_OFFSETOF(dir_config_t, s3_access_key),
    ACCESS_CONF,
    "Pass the Amazon S3 Access Key)"),

  AP_INIT_TAKE1("S3Region",
    ap_set_string_slot,
    (void*)APR_OFFSETOF(dir_config_t, s3_region),
    ACCESS_CONF,
    "Pass the Amazon S3 Region"),

  AP_INIT_TAKE1("S3SecurityToken",
    ap_set_string_slot,
    (void*)APR_OFFSETOF(dir_config_t, s3_security_token),
    ACCESS_CONF,
    "Pass the Amazon S3 Security Token"),

  AP_INIT_FLAG("S3UseHeaders",
    ap_set_flag_slot,
    (void*)APR_OFFSETOF(dir_config_t, s3_use_headers),
    ACCESS_CONF,
    "Set to 'on' to use headers for Amazon S3 authentication"),

  AP_INIT_TAKE1("TranscodeProxyPass",
    ap_set_string_slot,
    (void*)APR_OFFSETOF(dir_config_t, transcode_proxy_pass),
    ACCESS_CONF,
    "Pass transcoding to proxy"),

  AP_INIT_FLAG("UspPreferStatic",
    ap_set_flag_slot,
    (void*)APR_OFFSETOF(dir_config_t, usp_prefer_static),
    ACCESS_CONF,
    "Set to 'on' to prefer static files over dynamically generated files"),

  AP_INIT_FLAG("UspIssPassThrough",
    ap_set_flag_slot,
    (void*)APR_OFFSETOF(dir_config_t, usp_iss_pass_through),
    ACCESS_CONF,
    "Set to 'on' to pass through MP4 fragments directly to Smooth Streaming"),

  AP_INIT_FLAG("UspDynamicTimeShiftBufferDepth",
    ap_set_flag_slot,
    (void*)APR_OFFSETOF(dir_config_t, usp_dynamic_time_shift_buffer_depth),
    ACCESS_CONF,
    "Set to 'on' to dynamically adjust the @timeShiftBufferDepth"),

  AP_INIT_FLAG("UspSkipRewrite",
    ap_set_flag_slot,
    (void*)APR_OFFSETOF(dir_config_t, usp_skip_rewrite),
    ACCESS_CONF,
    "Set to 'off' to enable the internal rewrite rules"),

  AP_INIT_FLAG("UspEnableMMAP",
    ap_set_flag_slot,
    (void*)APR_OFFSETOF(dir_config_t, usp_enable_mmap),
    ACCESS_CONF,
    "Set to 'off' to disable use of memory-mapping"),

  AP_INIT_FLAG("UspEnableSubreq",
    ap_set_flag_slot,
    (void*)APR_OFFSETOF(dir_config_t, usp_enable_subreq),
    ACCESS_CONF,
    "Set to 'on' to enable use of subrequests"),

  // Smooth Streaming
  AP_INIT_FLAG("UspHandleIsm",
    ap_set_flag_slot,
    (void*)APR_OFFSETOF(dir_config_t, usp_handle_ism),
    ACCESS_CONF,
    "Set to 'on' to enable mod_smooth_streaming functionality"),

  AP_INIT_FLAG("UspHandleF4f",
    ap_set_flag_slot,
    (void*)APR_OFFSETOF(dir_config_t, usp_handle_f4f),
    ACCESS_CONF,
    "Set to 'on' to enable mod_f4fhttp functionality"),

  AP_INIT_FLAG("UspHandleApi",
    ap_set_flag_slot,
    (void*)APR_OFFSETOF(dir_config_t, usp_handle_api),
    ACCESS_CONF,
    "Set to 'on' to enable live publishing point api functionality"),

  AP_INIT_TAKE1("Fmp4IssFragmentNotFound",
    ap_set_int_slot,
    (void*)APR_OFFSETOF(dir_config_t, fmp4_iss_fragment_not_found),
    ACCESS_CONF,
    "HTTP code (404) returned when a fragment is requested before the DVR window (too old)"),

  AP_INIT_TAKE1("Fmp4IssFragmentMissing",
    deprecated_option,
    NULL,
    ACCESS_CONF,
    "Deprecated: HTTP code (412) returned when a fragment is missing in the DVR window"),

  AP_INIT_TAKE1("Fmp4IssFragmentNotAvailable",
    ap_set_int_slot,
    (void*)APR_OFFSETOF(dir_config_t, fmp4_iss_fragment_not_available),
    ACCESS_CONF,
    "HTTP code (412) returned when a fragment is requested after the DVR window (too new)"),

  // HTTP Dynamic Streaming
  AP_INIT_TAKE1("Fmp4HdsFragmentNotFound",
    ap_set_int_slot,
    (void*)APR_OFFSETOF(dir_config_t, fmp4_hds_fragment_not_found),
    ACCESS_CONF,
    "HTTP code (404) returned when a fragment is requested before the DVR window (too old)"),

  AP_INIT_TAKE1("Fmp4HdsFragmentMissing",
    deprecated_option,
    NULL,
    ACCESS_CONF,
    "Deprecated: HTTP code (503) returned when a fragment is missing in the DVR window"),

  AP_INIT_TAKE1("Fmp4HdsFragmentNotAvailable",
    ap_set_int_slot,
    (void*)APR_OFFSETOF(dir_config_t, fmp4_hds_fragment_not_available),
    ACCESS_CONF,
    "HTTP code (503) returned when a fragment is requested after the DVR window (too new)"),

  // HTTP Live Streaming
  AP_INIT_TAKE1("Fmp4HlsFragmentNotFound",
    ap_set_int_slot,
    (void*)APR_OFFSETOF(dir_config_t, fmp4_hls_fragment_not_found),
    ACCESS_CONF,
    "HTTP code (404) returned when a fragment is requested before the DVR window (too old)"),

  AP_INIT_TAKE1("Fmp4HlsFragmentMissing",
    deprecated_option,
    NULL,
    ACCESS_CONF,
    "Deprecated: HTTP code (503) returned when a fragment is missing in the DVR window"),

  AP_INIT_TAKE1("Fmp4HlsFragmentNotAvailable",
    ap_set_int_slot,
    (void*)APR_OFFSETOF(dir_config_t, fmp4_hls_fragment_not_available),
    ACCESS_CONF,
    "HTTP code (503) returned when a fragment is requested after the DVR window (too new)"),

  // MPEG DASH Streaming
  AP_INIT_TAKE1("Fmp4MpdFragmentNotFound",
    ap_set_int_slot,
    (void*)APR_OFFSETOF(dir_config_t, fmp4_mpd_fragment_not_found),
    ACCESS_CONF,
    "HTTP code (404) returned when a fragment is requested before the DVR window (too old)"),

  AP_INIT_TAKE1("Fmp4MpdFragmentMissing",
    deprecated_option,
    NULL,
    ACCESS_CONF,
    "Deprecated: HTTP code (503) returned when a fragment is missing in the DVR window"),

  AP_INIT_TAKE1("Fmp4MpdFragmentNotAvailable",
    ap_set_int_slot,
    (void*)APR_OFFSETOF(dir_config_t, fmp4_mpd_fragment_not_available),
    ACCESS_CONF,
    "HTTP code (503) returned when a fragment is requested after the DVR window (too new)"),

  AP_INIT_TAKE1("UspOutputFilter",
    set_output_filter,
    NULL,
    ACCESS_CONF,
    "Filter output through an external command"),

  AP_INIT_FLAG("UspForceContentLength",
    ap_set_flag_slot,
    (void*)APR_OFFSETOF(dir_config_t, usp_force_content_length),
    ACCESS_CONF,
    "Set to 'on' to force determining Content-Length even for dynamic content"),

  {NULL}
};


#ifdef __cplusplus
extern "C" {
#endif

#define usp_module_name smooth_streaming_module

module AP_MODULE_DECLARE_DATA usp_module_name =
{
  STANDARD20_MODULE_STUFF,
  create_dir_config,
  merge_dir_config,
  create_server_config,
  merge_server_config,
  smooth_streaming_cmds,
  register_hooks
};

#ifdef __cplusplus
} /* extern C definitions */
#endif

static void* create_server_config(apr_pool_t* pool, server_rec* svr)
{
  server_config_t* config = (server_config_t*)apr_pcalloc(pool, sizeof(server_config_t));
  config->aliases = apr_array_make(pool, 10, sizeof(proxy_alias_t));

  config->gctx = NULL;
  config->license = NULL;

  return config;
}

static int merge_int(int base, int add)
{
  return add == -1 ? base : add;
}

static char const* merge_str(char const* base, char const* add)
{
  return add == NULL ? base : add;
}

static void* merge_server_config(apr_pool_t* pool, void* basev, void* addv)
{
  server_config_t* base = basev;
  server_config_t* add = addv;
  server_config_t* res = apr_palloc(pool, sizeof(server_config_t));

  res->aliases = apr_array_append(pool, base->aliases, add->aliases);
  res->license = merge_str(base->license, add->license);

  return res;
}

static void* create_dir_config(apr_pool_t* pool, char* dummy)
{
  dir_config_t* config = (dir_config_t*)apr_pcalloc(pool, sizeof(dir_config_t));

  config->usp_enable_mmap = -1;
  config->usp_enable_subreq = -1;
  config->usp_prefer_static = -1;
  config->usp_iss_pass_through = -1;
  config->usp_dynamic_time_shift_buffer_depth = -1;
  config->usp_skip_rewrite = -1;
  config->usp_handle_ism = -1;
  config->usp_handle_f4f = -1;
  config->usp_handle_api = -1;

  config->fmp4_iss_fragment_not_found = -1;
  config->fmp4_iss_fragment_not_available = -1;

  config->fmp4_hds_fragment_not_found = -1;
  config->fmp4_hds_fragment_not_available = -1;

  config->fmp4_hls_fragment_not_found = -1;
  config->fmp4_hls_fragment_not_available = -1;

  config->fmp4_mpd_fragment_not_found = -1;
  config->fmp4_mpd_fragment_not_available = -1;

  config->s3_secret_key = NULL;
  config->s3_access_key = NULL;
  config->s3_region = NULL;
  config->s3_security_token = NULL;
  config->s3_use_headers = -1;

  config->transcode_proxy_pass = NULL;

  config->usp_output_filter = NULL;

  config->usp_force_content_length = -1;

  return config;
}

static void* merge_dir_config(apr_pool_t* pool, void* basev, void* addv)
{
  dir_config_t* base = basev;
  dir_config_t* add = addv;
  dir_config_t* res = apr_palloc(pool, sizeof(dir_config_t));

  res->usp_enable_mmap = merge_int(base->usp_enable_mmap, add->usp_enable_mmap);
  res->usp_enable_subreq = merge_int(base->usp_enable_subreq, add->usp_enable_subreq);
  res->usp_prefer_static = merge_int(base->usp_prefer_static, add->usp_prefer_static);
  res->usp_iss_pass_through = merge_int(base->usp_iss_pass_through, add->usp_iss_pass_through);
  res->usp_dynamic_time_shift_buffer_depth = merge_int(base->usp_dynamic_time_shift_buffer_depth, add->usp_dynamic_time_shift_buffer_depth);
  res->usp_skip_rewrite = merge_int(base->usp_skip_rewrite, add->usp_skip_rewrite);
  res->usp_handle_ism = merge_int(base->usp_handle_ism, add->usp_handle_ism);
  res->usp_handle_f4f = merge_int(base->usp_handle_f4f, add->usp_handle_f4f);
  res->usp_handle_api = merge_int(base->usp_handle_api, add->usp_handle_api);

  res->fmp4_iss_fragment_not_found = merge_int(base->fmp4_iss_fragment_not_found, add->fmp4_iss_fragment_not_found);
  res->fmp4_iss_fragment_not_available = merge_int(base->fmp4_iss_fragment_not_available, add->fmp4_iss_fragment_not_available);
  res->fmp4_hds_fragment_not_found = merge_int(base->fmp4_hds_fragment_not_found, add->fmp4_hds_fragment_not_found);
  res->fmp4_hds_fragment_not_available = merge_int(base->fmp4_hds_fragment_not_available, add->fmp4_hds_fragment_not_available);
  res->fmp4_hls_fragment_not_found = merge_int(base->fmp4_hls_fragment_not_found, add->fmp4_hls_fragment_not_found);
  res->fmp4_hls_fragment_not_available = merge_int(base->fmp4_hls_fragment_not_available, add->fmp4_hls_fragment_not_available);
  res->fmp4_mpd_fragment_not_found = merge_int(base->fmp4_mpd_fragment_not_found, add->fmp4_mpd_fragment_not_found);
  res->fmp4_mpd_fragment_not_available = merge_int(base->fmp4_mpd_fragment_not_available, add->fmp4_mpd_fragment_not_available);

  res->s3_secret_key = merge_str(base->s3_secret_key, add->s3_secret_key);
  res->s3_access_key = merge_str(base->s3_access_key, add->s3_access_key);
  res->s3_region = merge_str(base->s3_region, add->s3_region);
  res->s3_security_token = merge_str(base->s3_security_token, add->s3_security_token);
  res->s3_use_headers = merge_int(base->s3_use_headers, add->s3_use_headers);

  res->transcode_proxy_pass = merge_str(base->transcode_proxy_pass, add->transcode_proxy_pass);

  res->usp_output_filter = add->usp_output_filter == NULL ? base->usp_output_filter : add->usp_output_filter;

  res->usp_force_content_length = merge_int(base->usp_force_content_length, add->usp_force_content_length);

  return res;
}

static int remap_status(fmp4_result result, int http_status, dir_config_t const* conf)
{
  int new_status = -1;

  switch(result)
  {
    case FMP4_ISS_FRAGMENT_NOT_FOUND:
      new_status = conf->fmp4_iss_fragment_not_found;
      break;
    case FMP4_ISS_FRAGMENT_NOT_YET_AVAILABLE:
      new_status = conf->fmp4_iss_fragment_not_available;
      break;
    case FMP4_HDS_FRAGMENT_NOT_FOUND:
      new_status = conf->fmp4_hds_fragment_not_found;
      break;
    case FMP4_HDS_FRAGMENT_NOT_YET_AVAILABLE:
      new_status = conf->fmp4_hds_fragment_not_available;
      break;
    case FMP4_HLS_FRAGMENT_NOT_FOUND:
      new_status = conf->fmp4_hls_fragment_not_found;
      break;
    case FMP4_HLS_FRAGMENT_NOT_YET_AVAILABLE:
      new_status = conf->fmp4_hls_fragment_not_available;
      break;
    case FMP4_MPD_FRAGMENT_NOT_FOUND:
      new_status = conf->fmp4_mpd_fragment_not_found;
      break;
    case FMP4_MPD_FRAGMENT_NOT_YET_AVAILABLE:
      new_status = conf->fmp4_mpd_fragment_not_available;
      break;
    default:
      break;
  }

  return new_status != -1 ? new_status : http_status;
}

static apr_status_t cleanup_libfmp4(void* context)
{
  mp4_global_context_t* gctx = context;
  libfmp4_global_exit(gctx);

  return APR_SUCCESS;
}

static int smooth_streaming_post_config(apr_pool_t* p, apr_pool_t* plog,
                                        apr_pool_t* ptemp, server_rec* s)
{
  char const* src = "apache mod_smooth_streaming";
  char const* version = X_MOD_SMOOTH_STREAMING_VERSION;

  mp4_global_context_t* gctx = NULL;
  server_config_t* conf = NULL;
  char const* license = NULL;
  void* tmp = NULL;
  char const* key = "gctx";

  ap_add_version_component(p, "USP/" MODULE_VERSION);
  ap_add_version_component(p, "IISMS/4.0");

#if 1 // prevent double initialization
  apr_pool_userdata_get(&tmp, key, s->process->pool);
  if(tmp == NULL)
  {
    apr_pool_userdata_set((const void*)1, key, apr_pool_cleanup_null,
                          s->process->pool);
    return OK;
  }
#endif

  gctx = libfmp4_global_init();
  apr_pool_cleanup_register(p, (void*)gctx, cleanup_libfmp4,
                            apr_pool_cleanup_null);

  for (; s; s = s->next)
  {
    conf = (server_config_t*)
      ap_get_module_config(s->module_config, &usp_module_name);

    conf->gctx = gctx; // set pointer

    if(conf->license) // find license key
    {
      license = conf->license;
    }
  }

  char const* policy_check_result =
    libfmp4_load_license(conf->gctx, src, version, license);

  if(policy_check_result != NULL)
  {
    ap_log_error(APLOG_MARK, APLOG_CRIT, 0, s, "%s", policy_check_result);
  }
  else if(license)
  {
    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,
                 "License key found: %s", license);
  }

  for(int i=0; i < sizeof(deprecated_options)/sizeof(deprecated_options[0]); ++i)
  {
    if(deprecated_options[i].count)
    {
      ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,
                   "Use of DEPRECATED option %s (ignored)",
                   deprecated_options[i].name);
    }
  }
 
  return OK;
}

static fmp4_http_method_t get_http_method(request_rec* r)
{
  switch(r->method_number)
  {
  case M_GET:  return FMP4_HTTP_GET;
  case M_POST: return FMP4_HTTP_POST;
  case M_PUT: return FMP4_HTTP_PUT;
  case M_DELETE: return FMP4_HTTP_DELETE;
  default:
    return r->method_number;
  }
}

static int map_apr_status_to_http(apr_status_t status)
{
  // Can't use a switch because of the APR_STATUS_IS_xxx() macros
  if(status == APR_SUCCESS)
  {
    return HTTP_OK;
  }
  else if(APR_STATUS_IS_TIMEUP(status))
  {
    return HTTP_REQUEST_TIME_OUT;
  }
  else if(ap_is_HTTP_VALID_RESPONSE(status))
  {
    return status;
  }
  else
  {
    return HTTP_INTERNAL_SERVER_ERROR;
  }
}

static apr_status_t
remove_output_filter(ap_filter_t* next, ap_filter_rec_t const* filter)
{
  for(; next; next = next->next)
  {
    if(next->frec == filter)
    {
      ap_remove_output_filter(next);
      return APR_SUCCESS;
    }
  }

  return APR_NOTFOUND;
}

static int
drive_smooth_streaming_get(request_rec* r, dir_config_t const* dir_conf,
                           mp4_process_context_t* context)
{
  apr_status_t rv = APR_SUCCESS;
  int http_status;
  headers_t* headers;
  buckets_t* buckets;
  uint64_t content_length = 0;
  mp4_split_options_t const* options = context->options_;
  char const* file = mp4_split_options_get_file(options);

  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                "smooth_streaming_handler filename=%s file=%s",
                context->filename_, file);

  // If we are requesting the server manifest file, then it may an encoding tool
  // that is using the GET request to get the authorization setup for the POST
  // or it's simply checking that the publishing point exists.
  if(mp4_ends_with(context->filename_, ".isml"))
  {
    if(mp4_starts_with(file, "Streams("))
    {
      return HTTP_OK;
    }
  }

  if(strncmp(context->filename_, "ismproxy:", sizeof("ismproxy:") - 1) == 0)
  {
    context->filename_ += sizeof("ismproxy:") - 1;
  }
#if 0 // Don't rely on (stale) file metadata
  else
  {
    apr_finfo_t fi;
    rv = apr_stat(&fi, context->filename_, APR_FINFO_SIZE, r->pool);

    if(rv == APR_SUCCESS)
    {
      context->filesize_ = fi.size;
    }
  }
#endif

  http_status = mp4_process(context);

  headers = context->headers_;
  buckets = context->buckets_;

  if(headers->x_usp_header1[0])
  {
    apr_table_set(r->err_headers_out, "X-USP-Info1", headers->x_usp_header1);
  }
  if(headers->x_usp_header2[0])
  {
    apr_table_set(r->err_headers_out, "X-USP-Info2", headers->x_usp_header2);
  }

  // Check if we already know the content length.
  content_length = buckets_size(buckets);
  if(content_length != UINT64_MAX)
  {
    TRACE1_R(r, "content_length is %" APR_UINT64_T_FMT
                ", no need to establish buckets size", content_length);
  }
  else if(dir_conf->usp_force_content_length == 1 ||
          (context->output_filter_ != NULL && r->proto_num < HTTP_VERSION(1,1)))
  {
    // In these special cases:
    // 1) UspForceContentLength is on, or
    // 2) Using an output_filter in combination with HTTP/1.0, such as with
    //    ApacheBench
    // establish the buckets size now, to get the actual content length.
    // NOTE: this can possibly read large amounts of data into memory.
    TRACE1_R(r, "establishing buckets size due to %s",
                dir_conf->usp_force_content_length == 1 ?
                  "UspForceContentLength" : "output_filter with HTTP/1.0");
    fmp4_result result = buckets_establish_size(buckets, &content_length);
    if(result != FMP4_OK)
    {
      ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                    "smooth_streaming_handler: failed to establish buckets "
                    "size (%s)", fmp4_result_to_string(result));
      return HTTP_INTERNAL_SERVER_ERROR;
    }
    TRACE1_R(r, "established buckets size: %" APR_UINT64_T_FMT, content_length);
  }
  else
  {
    TRACE1_R(r, "content_length is unknown, but not going to establish buckets "
                "size");
  }

  // If we know the content length then we can pass it in the header.
  if(content_length != UINT64_MAX)
  {
    ap_set_content_length(r, content_length);

    // Allow byte range requests. The ap_byterange_filter only works on
    // brigades containing an EOS and no buckets with unknown length.
    apr_table_setn(r->headers_out, "Accept-Ranges", "bytes");
  }

  // Add Last-Modified
  if(headers->updated_at_)
  {
    ap_update_mtime(r, headers->updated_at_);
    ap_set_last_modified(r);
  }

  typedef uint64_t seconds_t;

  // Add X-USP-Last-Modified
  if(headers->updated_at_experimental_)
  {
    seconds_t last_modified = apr_time_sec(headers->updated_at_experimental_);
    char* timestr = apr_palloc(r->pool, APR_RFC822_DATE_LEN);
    apr_rfc822_date(timestr, apr_time_from_sec(last_modified));
    apr_table_setn(r->headers_out, "X-USP-Last-Modified", timestr);
  }

  // Add Cache-Control and Expires
  if(headers->cache_control_)
  {
    apr_table_setn(r->headers_out, "Cache-Control", headers->cache_control_);
  }
  else if(headers->expires_at_)
  {
    // compute expiry with seconds accuracy for interoperable results with nginx
    seconds_t expires_time = apr_time_sec(headers->expires_at_);
    seconds_t request_time = apr_time_sec(r->request_time);

    // use expiry from backend only if it's at least a second ahead of request
    // otherwise, we skip insertion of Expires/Cache-Control and leave this up
    // to the customer (see: mod_headers, mod_expires)
    if(expires_time < request_time + 1)
    {
      expires_time = request_time + 1;
    }

    seconds_t max_age = expires_time - request_time;
    apr_table_setn(r->headers_out, "Cache-Control",
                   apr_psprintf(r->pool, "max-age=%" APR_UINT64_T_FMT, max_age));

    char* timestr = apr_palloc(r->pool, APR_RFC822_DATE_LEN);
    apr_rfc822_date(timestr, apr_time_from_sec(expires_time));
    apr_table_setn(r->headers_out, "Expires", timestr);
  }

  // Add ETag
  if(headers->cache_control_ != cache_control_no_cache)
  {
    if(headers->etag_[0])
    {
      apr_table_setn(r->headers_out, "ETag", headers->etag_);
    }
    else
    {
      // The ETag takes into account the last modified time (r->mtime). Since
      // for dynamically generated content we set the (high resolution)
      // headers->updated_at_ as the last modified time, we can simply
      // generate an ETag and do nothing special.
      ap_set_etag(r);
    }
  }

  if(headers->sunset_at_)
  {
    seconds_t sunset_time = apr_time_sec(headers->sunset_at_);
    char* timestr = apr_palloc(r->pool, APR_RFC822_DATE_LEN);
    apr_rfc822_date(timestr, apr_time_from_sec(sunset_time));
    apr_table_setn(r->headers_out, "Sunset", timestr);
  }

  // Add Link: header for start/next
  if(headers->link_[0])
  {
    apr_table_setn(r->headers_out, "Link", headers->link_);
  }

  // Add Location
  if(headers->location_)
  {
    apr_table_setn(r->headers_out, "Location", headers->location_);
  }

  if(http_status != HTTP_OK)
  {
    if(!context->is_rest_api_)
    {
      http_status = remap_status(context->result_, http_status, dir_conf);
    }

    r->status = http_status;

    return OK;
  }

  // Check for conditional requests
  {
    int errstatus;
    if((errstatus = ap_meets_conditions(r)) != OK)
    {
      // For the HTTP_NOT_MODIFIED case set r->status and return OK in order to
      // prevent running through the error processing stack.
      if(errstatus == HTTP_NOT_MODIFIED)
      {
        r->status = errstatus;
        return OK;
      }

      return errstatus;
    }
  }

  // Set the content type
  ap_set_content_type(r, headers->content_type_);

  if(!r->header_only)
  {
    apr_bucket_brigade* bb;
    bucket_t* head = buckets->bucket_;
    bucket_t* bucket = bucket_next(head);

    if(content_length == UINT64_MAX)
    {
      if(r->proto_num >= HTTP_VERSION(1,1))
      {
        // if we are transforming the buckets dynamically, and we don't know the
        // size yet, so switch to "Transfer-Encoding: chunked".
        r->chunked = 1;

        // we have to remove the core content-length filter. It will never set
        // the Content-Length anyways, but it always iterates over the complete
        // brigate (and as a side-effect evaluates all the xfrm buckets which
        // may read large amounts).
        remove_output_filter(r->output_filters, ap_content_length_filter_handle);
      }
      else
      {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                      "smooth_streaming_handler: unknown content length with "
                      "HTTP < 1.1, failing request");

        return HTTP_VERSION_NOT_SUPPORTED;
      }
    }

    bb = apr_brigade_create(r->pool, r->connection->bucket_alloc);

    while(bucket != head)
    {
      apr_bucket* e;

      bucket_t* next = bucket_next(bucket);
      bucket_remove(bucket);
      e = apr_bucket_usp_create(bucket, bb->bucket_alloc);
      APR_BRIGADE_INSERT_TAIL(bb, e);

      bucket = next;
    }

    // Add EOS bucket
    APR_BRIGADE_INSERT_TAIL(bb, apr_bucket_eos_create(bb->bucket_alloc));

    TRACE1_R(r, "calling ap_pass_brigade");
    rv = ap_pass_brigade(r->output_filters, bb);
    TRACE1_R(r, "called ap_pass_brigade, rv=%d", rv);

    if(r->connection->aborted)
    {
      ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                    "smooth_streaming_handler: connection aborted, rv=%d", rv);
      return HTTP_INTERNAL_SERVER_ERROR;
    }

    if(rv != APR_SUCCESS)
    {
      char errmsg[120];
      apr_strerror(rv, errmsg, sizeof errmsg);
      ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                    "smooth_streaming_handler: ap_pass_brigade failed: %s",
                    errmsg);

      if(r->chunked)
      {
        // Assuming the status line has already been sent, we force a protocol
        // error to ensure the client at least notices *some* kind of failure.
        // See modules/http/chunk_filter.c in the Apache source.
        //
        // With curl as a client, this results in an error "curl: (18) transfer
        // closed with outstanding read data remaining"

        apr_brigade_cleanup(bb);
        APR_BRIGADE_INSERT_TAIL(bb, ap_bucket_error_create(
          HTTP_BAD_GATEWAY, NULL, r->pool, bb->bucket_alloc));
        APR_BRIGADE_INSERT_TAIL(bb, apr_bucket_eos_create(bb->bucket_alloc));

        TRACE1_R(r, "calling ap_pass_brigade with error bucket");
        rv = ap_pass_brigade(r->output_filters, bb);
        TRACE1_R(r, "called ap_pass_brigade with error bucket, rv=%d", rv);

        // We need to return something else than HTTP_INTERNAL_SERVER_ERROR or
        // OK here: with the former, Apache will helpfully append a "the server
        // has encountered an internal error" document to the output, and with
        // the latter, the access log will pretend that the request was fine
        // (http status 200).
        return DONE;
      }

      return map_apr_status_to_http(rv);
    }
  }

  return OK;
}

static int
drive_smooth_streaming_post(request_rec* r, mp4_process_context_t* context)
{
  apr_size_t chunk_size = 8 * 1024;
  mp4_split_options_t const* options = context->options_;
  char const* file = mp4_split_options_get_file(options);

  if(!mp4_ends_with(context->filename_, ".isml") && !context->is_rest_api_)
  {
    return HTTP_FORBIDDEN;
  }

  // Connect
  if(!file[0] && !context->is_rest_api_)
  {
    return 0;
  }

  // Pushing to Streams(...), disable api
  if(file[0] && context->is_rest_api_)
  {
    context->is_rest_api_ = 0;
  }

  if(strncmp(context->filename_, "ismproxy:", sizeof("ismproxy:") - 1) == 0)
  {
    context->filename_ += sizeof("ismproxy:") - 1;
  }

  if(!strcmp(file, "purge"))
  {
    int http_status = mp4_process(context);

    if(http_status != HTTP_OK)
    {
      return http_status;
    }

    return OK;
  }

  {
    int http_status = HTTP_OK;
    char* buf = apr_palloc(r->pool, chunk_size);

    apr_bucket_brigade* bb =
      apr_brigade_create(r->pool, r->connection->bucket_alloc);

    post_handler_t* post_handler = create_post_handler(context);

    if(!post_handler)
    {
      ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                    "usp_post_handler returned: %s", context->result_text_);

      http_status = HTTP_INTERNAL_SERVER_ERROR;
    }

    if(http_status == HTTP_OK)
    {
      for(;;)
      {
        apr_size_t len = chunk_size;

        // If you get an error similar to:
        // (70007)The timeout specified has expired: Error reading chunk this
        // means that we didn't receive any data for the timeout specified.
        // Note that the timeout value may also be set by mod_reqtimeout and
        // the default setting may be quite agressive (10 seconds).
        // Note also that setting the RequestReadTimeout in a virtual host to
        // override the global settings, does *not* work. You have to set the
        // global RequestReadTimeout to a larger timeout value, or disable it
        // completely with:
        //   RequestReadTimeout header=0 body=0
        //
        // Also make sure to include in your (vhost) configuration the
        // following:
        //   LimitRequestBody 0

        apr_status_t status = ap_get_brigade(r->input_filters, bb,
                                             AP_MODE_READBYTES,
                                             APR_BLOCK_READ, len);

        if(status != APR_SUCCESS)
        {
          apr_brigade_destroy(bb);
          ap_log_rerror(APLOG_MARK, APLOG_ERR, status, r,
                        "mod_smooth_streaming(%s): ap_get_brigade",
                        file);
          http_status = map_apr_status_to_http(status);
          break;
        }

        status = apr_brigade_flatten(bb, buf, &len);
        if(status != APR_SUCCESS)
        {
          apr_brigade_destroy(bb);
          ap_log_rerror(APLOG_MARK, APLOG_ERR, status, r,
                        "mod_smooth_streaming(%s): apr_brigade_flatten",
                        file);
          http_status = map_apr_status_to_http(status);
          break;
        }

        apr_brigade_cleanup(bb);

#if 0 && defined(_DEBUG)
        printf("mod_smooth_streaming.c(%s): len=%u\n",
               r->args ? r->args : "", (unsigned int)len);
#endif

#if 1
        char result_text[256] = { 0 };

        if(post_handler)
        {
          ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
              "ingest stream %s (%lu bytes)", file, len);
          http_status = post_handler_insert(post_handler,
            (unsigned char*)buf, (unsigned char*)buf + len,
            result_text, sizeof(result_text) / sizeof(result_text[0]));
        }

        if(!ap_is_HTTP_SUCCESS(http_status))
        {
          if(result_text[0])
          {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
              "usp_post_handler(%s) returned: %u %s",
              file,
              http_status, result_text);
          }

          // abort the connection in case we are receiving the POST data as
          // chunked. Unfortunately we lose the original http_status
          // code and Apache always returns a 500 Internal Server Error.
          if(len)
          {
            http_status = HTTP_INTERNAL_SERVER_ERROR;
          }

          break;
        }
#endif

        if(len == 0)
        {
          break;
        }
      }
    }

    if(post_handler)
    {
      post_handler_exit(post_handler);
    }

#if 0 && defined(_DEBUG)
    printf("mod_smooth_streaming.c(%s): http=%d\n", file, http_status);
#endif

    return http_status == HTTP_OK ? OK : http_status;
  }
}

static int handle_ism_proxy(request_rec* r, int apply)
{
  // check or apply ism_proxy_pass
  const char* prefix = apply ? "apply" : "check";
  server_config_t* conf = (server_config_t*)
    ap_get_module_config(r->server->module_config, &usp_module_name);
  int i;
  proxy_alias_t* ent = (proxy_alias_t*) conf->aliases->elts;
  for (i = 0; i < conf->aliases->nelts; i++)
  {
    const char* fake = ent[i].fake;
    const char* real = ent[i].real;
    int in_location = ent[i].in_location;

    DEBUG_R(r, "%s_ism_proxy(step1) fake=%s real=%s in_location=%d uri=%s "
      "filename=%s", prefix, fake, real, in_location, r->uri, r->filename);

    // filename: /var/www/test/ism_proxy_pass.ssm/ism proxy pass.ism
    //     fake: /var/www/test/ism_proxy_pass.ssm/
    // OR
    //      uri: /ism_proxy_pass.ssm/ism proxy pass.ism
    //     fake: /ism_proxy_pass.ssm/
    //
    //     real: http://localhost:81/test/ism_proxy_pass.ssm/
    //    final: ismproxy:http://localhost:81/test/ism_proxy_pass.ssm/ism%20proxy%20pass.ism
    char const* to_match = in_location ? r->uri : r->filename;
    if(mp4_starts_with(to_match, fake))
    {
      char* escaped = ap_escape_uri(r->pool, to_match + strlen(fake));
      char* final = apr_psprintf(r->pool, "ismproxy:%s%s", real, escaped);
      DEBUG_R(r, "%s_ism_proxy(step2) final=%s", prefix, final);
      if(apply)
      {
        r->filename = final;
      }
      return OK;
    }
  }

  TRACE1_R(r, "%s_ism_proxy: declined", prefix);
  return DECLINED;
}

static int smooth_streaming_translate_name(request_rec* r)
{
  int res = OK;

  dir_config_t* conf = (dir_config_t*)
    ap_get_module_config(r->per_dir_config, &usp_module_name);

  int usp_handle_ism = merge_int(0, conf->usp_handle_ism);
  int usp_handle_f4f = merge_int(0, conf->usp_handle_f4f);
  int usp_handle_api = merge_int(0, conf->usp_handle_api);

  if(usp_handle_api && !(usp_handle_ism || usp_handle_f4f))
  {
    r->handler = SMOOTH_STREAMING_HANDLER;
    TRACE1_R(r, "translate_name: call core");
    return ap_core_translate(r);
  }

  // Check if we are enabled for this location
  if(!(usp_handle_ism || usp_handle_f4f))
  {
    TRACE1_R(r, "translate_name: declined, neither ism nor f4f handler");
    return DECLINED;
  }

  DEBUG_R(r, "translate(step1) hostname=%s uri=%s args=%s filename=%s",
    r->hostname, r->uri, r->args, r->filename);

  int usp_skip_rewrite = merge_int(1, conf->usp_skip_rewrite);

  char new_path[4096];
  char new_args[4096];

  char const* path_first = r->uri;
  char const* path_last = r->uri + strlen(r->uri);
  char const* query_first = r->args;
  char const* query_last = r->args + (r->args ? strlen(r->args) : 0);

  // mp4_rewrite_url returns 1 when the request needs to be handled by us
  TRACE1_R(r, "translate_name: calling mp4_rewrite_url, path=%s, query=%s,"
    " usp_handle_ism=%d, usp_handle_f4f=%d", r->uri, r->args, usp_handle_ism,
    usp_handle_f4f);
  if(!mp4_rewrite_url(path_first, path_last,
                      query_first, query_last,
                      new_path, sizeof(new_path),
                      new_args, sizeof(new_args),
                      usp_handle_ism, usp_handle_f4f))
  {
    TRACE1_R(r, "translate_name: declined, no rewrite done");
    res = DECLINED;
  }
  else
  {
    TRACE1_R(r, "translate_name: called mp4_rewrite_url, new path=%s, args=%s",
      new_path, new_args);

    // from here on out, our main handler should pick this up.
    r->handler = SMOOTH_STREAMING_HANDLER;

    TRACE1_R(r, "translate_name: usp_skip_rewrite=%d, usp_handle_f4f=%d",
      usp_skip_rewrite, usp_handle_f4f);
    if(usp_skip_rewrite && !usp_handle_f4f)
    {
      res = ap_core_translate(r);
      if(res != OK || !r->filename)
      {
        TRACE1_R(r, "translate_name: core translate failed, return %d", res);
        return res;
      }
      // Now it should be safe to check for IsmProxyPass directives.
      int use_ism_proxy = handle_ism_proxy(r, 0);
      TRACE1_R(r, "translate_name: use_ism_proxy=%d", use_ism_proxy);
      if(use_ism_proxy != OK)
      {
        TRACE1_R(r, "translate_name: declined, rewrite skipped");
        r->args = apr_psprintf(r->pool, "%s", new_args);
        return DECLINED;
      }
    }

    // store the rewritten url
    r->uri = apr_psprintf(r->pool, "%s", new_path);
    r->args = apr_psprintf(r->pool, "%s", new_args);
    r->filename = NULL;

    TRACE1_R(r, "translate_name: rewrote uri=%s args=%s", r->uri, r->args);
    res = OK;
  }
  //assert(res == OK || res == DECLINED);

  if(usp_handle_ism)
  {
    res = ap_core_translate(r);
    if(res != OK || !r->filename)
    {
      TRACE1_R(r, "translate_name: core translate failed, return %d",
        res);
      return res;
    }

    DEBUG_R(r, "translate(step2) hostname=%s uri=%s args=%s filename=%s",
      r->hostname, r->uri, r->args, r->filename);

    if((res = handle_ism_proxy(r, 1)) == OK)
    {
      // if the name was rewritten by ism_proxy_pass, then return
      TRACE1_R(r, "translate_name: handled by ism_proxy, filename=%s",
        r->filename);
      return OK;
    }
  }

  // otherwise allow other modules to run their hooks, e.g. mod_auth_token
  TRACE1_R(r, "translate_name: declined");
  return DECLINED;
}

static int smooth_streaming_map_location(request_rec* r)
{
  DEBUG_R(r, "map_to_storage(step1) filename=%s", r->filename);

  if(!r->filename || strncmp(r->filename, "ismproxy:", sizeof("ismproxy:") - 1) != 0)
  {
    TRACE1_R(r, "map_to_storage: declined");
    return DECLINED;
  }

  DEBUG_R(r, "map_to_storage(step2) filename=%s", r->filename);

  return OK;
}

static void set_header_in(void* p, char const* key, char const* value)
{
  request_rec* r = (request_rec*)p;
  DEBUG_R(r, "set header=%s, value=%s", key, value);
  apr_table_set(r->headers_in, key, value);
}

static int smooth_streaming_fixups(request_rec* r)
{
  DEBUG_R(r, "fixups, subreq=%s hostname=%s uri=%s filename=%s",
    (r->main == NULL ? "no" : "yes"), r->hostname, r->uri, r->filename);

  if(r->main == NULL)
  {
    DEBUG_R(r, "Not adding S3 headers for main request");
    return DECLINED;
  }

  if(r->filename == NULL || strncmp(r->filename, "proxy:", 6) != 0)
  {
    DEBUG_R(r, "Not adding S3 headers for non-proxy request");
    return DECLINED;
  }

  char const* url = r->filename + 6;
  DEBUG_R(r, "url=%s", url);

  dir_config_t* dir_conf =
    ap_get_module_config(r->per_dir_config, &usp_module_name);
  char const* src = "sub";
  if(dir_conf->s3_secret_key == NULL || dir_conf->s3_access_key == NULL)
  {
    // If the subrequest does not have any S3 parameters, try again with the
    // configuration of the main request.
    dir_conf = ap_get_module_config(r->main->per_dir_config, &usp_module_name);
    if(dir_conf->s3_secret_key == NULL || dir_conf->s3_access_key == NULL)
    {
      DEBUG_R(r, "Not adding S3 headers, no secret or access keys");
      return DECLINED;
    }
    src = "main";
  }

  if(dir_conf->s3_use_headers <= 0)
  {
    DEBUG_R(r, "Not adding S3 headers, S3UseHeaders is off");
    return DECLINED;
  }

  DEBUG_R(r, "Adding S3 headers, src=%s, url=%s, region=%s",
    src, url, dir_conf->s3_region);

  char result_text[256] = { 0 };
  int status = mp4_add_s3_headers(
    r, url, dir_conf->s3_secret_key, dir_conf->s3_access_key,
    dir_conf->s3_region, dir_conf->s3_security_token, set_header_in,
    result_text, sizeof result_text);

  if(status != 200)
  {
    ERROR_R(r, "adding S3 signature headers failed with status %d: %s", status,
      result_text);
    return status;
  }

#if 0
  if(apr_table_get(r->headers_in, "X-USP-S3-Signed") == NULL)
  {
    apr_table_set(r->headers_in, "X-USP-S3-Signed", "yes");
  }
#endif

  return DECLINED;
}

static int get_server_variable(void* context, char const* name, char const** value)
{
  request_rec* r = (request_rec*)context;

  if(!strcmp(name, "REQUEST_SCHEME"))
  {
    *value = ap_http_scheme(r);
    return (int)strlen(*value);
  } else
  if(!strcmp(name, "HTTP_HOST"))
  {
    *value = apr_table_get(r->headers_in, "Host");
    return (int)strlen(*value);
  } else
  if(!strcmp(name, "REQUEST_URI"))
  {
    *value = r->uri;
    return (int)strlen(*value);
  }

  return 0;
}

static fmp4_log_level_t apache_log_level_to_fmp4_log_level(request_rec* r)
{
#if defined(APLOG_R_IS_LEVEL) // Apache 2.2 does not have APLOG_R_IS_LEVEL
  if(APLOG_R_IS_LEVEL(r, APLOG_TRACE1))
  {
    return FMP4_LOG_TRACE;
  }
  if(APLOG_R_IS_LEVEL(r, APLOG_DEBUG))
  {
    return FMP4_LOG_DEBUG;
  }
  if(APLOG_R_IS_LEVEL(r, APLOG_NOTICE))
  {
    return FMP4_LOG_DEBUG;
  }
  if(APLOG_R_IS_LEVEL(r, APLOG_INFO))
  {
    return FMP4_LOG_INFO;
  }
  if(APLOG_R_IS_LEVEL(r, APLOG_WARNING))
  {
    return FMP4_LOG_WARNING;
  }

  return FMP4_LOG_ERROR;
#else
  return FMP4_LOG_DEBUG;
#endif
}

static void log_error_callback(void* context, fmp4_log_level_t level,
                               char const* str)
{
  request_rec* r = (request_rec*)context;

  int log_level;
  switch(level)
  {
  case FMP4_LOG_TRACE:
#if defined(APLOG_TRACE1) // Apache 2.2 does not have trace, available since 2.4
    log_level = APLOG_TRACE1;
    break;
#endif
  case FMP4_LOG_DEBUG:
    log_level = APLOG_DEBUG;
    break;
  case FMP4_LOG_INFO:
    log_level = APLOG_INFO;
    break;
  case FMP4_LOG_NOTICE_DO_NOT_USE:
    log_level = APLOG_NOTICE;
    break;
  case FMP4_LOG_WARNING:
    log_level = APLOG_WARNING;
    break;
  case FMP4_LOG_ERROR:
  default:
    log_level = APLOG_ERR;
    break;
  }

  ap_log_rerror(APLOG_MARK, log_level, 0, r, "%s", str);
}

static apr_status_t mp4_process_context_cleanup(void* data)
{
  mp4_process_context_t* context = data;

  mp4_process_context_exit(context);

  return APR_SUCCESS;
}

static int smooth_streaming_handler(request_rec* r)
{
  int http_status;

  if((!r->handler) || (strcmp(r->handler, SMOOTH_STREAMING_HANDLER)))
  {
    return DECLINED;
  }

  server_config_t const* server_conf =
    ap_get_module_config(r->server->module_config, &usp_module_name);

  dir_config_t const* dir_conf =
    ap_get_module_config(r->per_dir_config, &usp_module_name);

  r->allowed = (AP_METHOD_BIT << M_GET);
  r->allowed |= (AP_METHOD_BIT << M_POST);
  r->allowed |= (AP_METHOD_BIT << M_PUT);
  if(dir_conf->usp_handle_api > 0)
  {
    r->allowed |= (AP_METHOD_BIT << M_DELETE);
  }

  // let the default Apache handler deal with the OPTIONS method.
  if(r->method_number == M_OPTIONS)
  {
    return DECLINED;
  }

  http_status = HTTP_METHOD_NOT_ALLOWED;

  if((AP_METHOD_BIT << r->method_number) & r->allowed)
  {
    mp4_process_context_t* context = apr_palloc(r->pool,
                                                sizeof(mp4_process_context_t));

    mp4_process_context_init(context, server_conf->gctx);

    apr_pool_cleanup_register(r->pool, context, mp4_process_context_cleanup,
                              apr_pool_cleanup_null);

    ap_set_module_config(r->request_config, &usp_module_name, context);

    context->filename_ = r->filename;
    context->verbose_ = apache_log_level_to_fmp4_log_level(r);

    if(dir_conf->usp_enable_subreq > 0)
    {
      TRACE1_R(r, "enabling download via subrequest, hostname=%s uri=%s, "
        "filename=%s args=%s", r->hostname, r->uri, r->filename, r->args);
      context->download_ = subreq_download;
      context->download_context_ = r;
    }

    context->get_server_variable_callback_ = get_server_variable;
    context->get_server_variable_context_ = r;
    context->log_error_callback_ = log_error_callback;
    context->log_error_context_ = r;
    context->method_ = get_http_method(r);

    {
      context->prefer_static_ = merge_int(0, dir_conf->usp_prefer_static);
      context->iss_pass_through_ = merge_int(0, dir_conf->usp_iss_pass_through);
      context->dynamic_time_shift_buffer_depth_ = merge_int(0, dir_conf->usp_dynamic_time_shift_buffer_depth);

      if(!dir_conf->usp_enable_mmap)
      {
        context->create_handler_io_ = create_handler_io_no_mmap;
      }

      context->s3_secret_key_ = dir_conf->s3_secret_key;
      context->s3_access_key_ = dir_conf->s3_access_key;
      context->s3_region_ = dir_conf->s3_region;
      context->s3_security_token_ = dir_conf->s3_security_token;
      context->s3_use_headers_ = merge_int(0, dir_conf->s3_use_headers);
      context->transcode_proxy_pass_ = dir_conf->transcode_proxy_pass;

      if(dir_conf->usp_handle_api > 0)
      {
        //if(r->method_number == M_POST)
        //{
        //  // if its resolved correctly, there is already an existing dir
        //  if(strlen(r->path_info) == 0)
        //  {
        //    return HTTP_CONFLICT;
        //  }

        //  // Check that there is no subdirs in path info
        //  if(strchr(r->path_info + 1, '/') != NULL)
        //  {
        //    return HTTP_BAD_REQUEST;
        //  }
        //}

        // apache stops translation on non existent dir,
        // restore the full path by joining walked path + unwalked path_info
        context->filename_ = apr_psprintf(r->pool, "%s%s", r->filename,
                                          r->path_info);

        context->is_rest_api_ = 1;
      }
    }

    // Module version info
    apr_table_setn(r->headers_out, "X-USP", fmp4_version_string());

    if(r->args && !mp4_split_options_set(context->options_, r->args,
                                         (unsigned int)strlen(r->args)))
    {
      return HTTP_BAD_REQUEST;
    }

    switch(r->method_number)
    {
    case M_GET:
      if(dir_conf->usp_output_filter != NULL)
      {
        char const* err;
        char const* val = ap_expr_str_exec(r, dir_conf->usp_output_filter,
                                           &err);
        if(err != NULL)
        {
          ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                        "smooth_streaming_handler: failed to evaluate "
                        "UspOutputFilter value expression: %s", err);
          r->status = HTTP_INTERNAL_SERVER_ERROR;
          return r->status;
        }
        TRACE1_R(r, "evaluated output_filter to %s", val);
        context->output_filter_ = val;
      }
      http_status = drive_smooth_streaming_get(r, dir_conf, context);
      break;
    case M_POST:
    case M_PUT:
      http_status = drive_smooth_streaming_post(r, context);
      break;
    case M_DELETE:
      if(context->is_rest_api_)
      {
        http_status = mp4_process(context);
      }
      break;
    }

    if(context->result_ != FMP4_OK)
    {
      mp4_split_options_t const* options = context->options_;
      char const* file = mp4_split_options_get_file(options);

      ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                    "mp4_process(uri %s, file %s) returned: %s %s",
                    context->filename_,
                    file,
                    fmp4_result_to_string(context->result_),
                    context->result_text_);
    }
  }

  if(http_status != OK)
  {
    TRACE1_R(r, "set request status to %d", http_status);
    r->status = http_status;
  }

  return ap_is_HTTP_SUCCESS(http_status) ? OK : http_status;
}

static char const* set_usp_license_key(cmd_parms* cmd, void* cfg, char const* arg)
{
  server_rec* s = cmd->server;
  server_config_t* conf = (server_config_t*) ap_get_module_config(s->module_config, &usp_module_name);

  conf->license = arg;

  return NULL;
}

static int
in_section(ap_directive_t const *dirp, char const* name1, char const* name2)
{
  while(dirp->parent != NULL)
  {
    dirp = dirp->parent;

    if(strcasecmp(dirp->directive, name1) == 0 ||
       strcasecmp(dirp->directive, name2) == 0)
    {
      return 1;
    }
  }

  return 0;
}

static char const* set_ism_proxy_pass(cmd_parms* cmd, void* cfg, char const* arg1, char const* arg2)
{
  server_rec* s = cmd->server;
  server_config_t* conf = (server_config_t*) ap_get_module_config(s->module_config, &usp_module_name);
  char const* fake = cmd->path;
  char const* real = NULL;

  int in_location = in_section(cmd->directive, "<Location", "<LocationMatch");

  if(fake == NULL)
  {
    // Outside a Directory or Location section, two arguments (fake and real)
    // are required.
    if(arg2 == NULL)
    {
      return "IsmProxyPass needs two arguments (virtual path and real url) "
             "when used outside a Directory or Location section";
    }
    fake = arg1;
    real = arg2;
  }
  else
  {
    // Inside a Directory or Location section, one argument (real) is required.
    if(arg2 != NULL)
    {
      return "IsmProxyPass needs only one argument (real url) when used inside "
             "a Directory or Location section";
    }
    real = arg1;
  }

  proxy_alias_t* alias = apr_array_push(conf->aliases);
  alias->fake = apr_pstrdup(cmd->pool, fake);
  alias->real = apr_pstrdup(cmd->pool, real);
  alias->in_location = in_location;

  return NULL;
}

static char const* set_output_filter(cmd_parms* cmd, void* cfg, char const* arg)
{
  char const* err;
  ap_expr_info_t const* expr =
    ap_expr_parse_cmd(cmd, arg, AP_EXPR_FLAG_STRING_RESULT, &err, NULL);
  if(err != NULL)
  {
    return apr_pstrcat(cmd->temp_pool, "Failed to parse UspOutputFilter "
                       "expression: ", err, NULL);
  }

  dir_config_t* dir_cfg = cfg;
  dir_cfg->usp_output_filter = expr;
  return NULL;
}

static void register_hooks(apr_pool_t* p)
{
  // NOTE: mod_auth_token has mod_alias as its successor, so we need to have
  // that too. Otherwise, if mod_auth_token is not loaded, we will not run
  // before mod_alias, and that conflicts with our translate_name processing.
  static const char* const we_run_before[] = { "mod_alias.c",
                                               "mod_auth_token.c", NULL };

  /* module initializer */
  ap_hook_post_config(smooth_streaming_post_config, NULL, NULL,
                      APR_HOOK_MIDDLE);

  subreq_init();

  /* URI-to-filename translation */
  ap_hook_translate_name(smooth_streaming_translate_name, NULL, we_run_before,
                         APR_HOOK_MIDDLE);
  /* storage mapping */
  ap_hook_map_to_storage(smooth_streaming_map_location, NULL, NULL,
                         APR_HOOK_FIRST);

  /* at end of request preparation, but before the handler */
  ap_hook_fixups(smooth_streaming_fixups, NULL, NULL, APR_HOOK_LAST);

  /* the handler */
  ap_hook_handler(smooth_streaming_handler, NULL, NULL, APR_HOOK_MIDDLE);

}

// End Of File

